package au.edu.adelaide.cs.condor;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Reader;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.apache.commons.csv.CSVFormat;
import org.apache.commons.csv.CSVRecord;

public class TextualFeaturesToNumeric {

	public static void main(String[] args) throws IOException {
		Set<String> stopwords = loadStopwords("stopwords.txt");
		PorterStemmer stemmer = new PorterStemmer();
		Map<String, Integer> tagFreq = new HashMap<String, Integer>();
		Map<String, Integer> titleFreq = new HashMap<String, Integer>();
		Map<String, Integer> textFreq = new HashMap<String, Integer>();
		Map<String, Integer> codeFreq = new HashMap<String, Integer>();
		Reader in = new FileReader("merged_data.csv");
		Iterable<CSVRecord> records = CSVFormat.EXCEL.withFirstRecordAsHeader().parse(in);
		List<String> headers = new ArrayList<String>();
		for (CSVRecord record : records) {
			headers = new ArrayList<String>(record.toMap().keySet());
			updateFrequencies(tagFreq, record, "Tags", stopwords, stemmer);
			updateFrequencies(titleFreq, record, "Title", stopwords, stemmer);
			updateFrequencies(textFreq, record, "ContentText", stopwords, stemmer);
			updateFrequencies(codeFreq, record, "ContentCode", stopwords, stemmer);
		}
		Map<String, Integer> tagFreqSorted = sortByValue(tagFreq);
		Map<String, Integer> titleFreqSorted = sortByValue(titleFreq);
		Map<String, Integer> textFreqSorted = sortByValue(textFreq);
		Map<String, Integer> codeFreqSorted = sortByValue(codeFreq);
		tagFreqSorted = limitToMostFrequent(tagFreq, 5);
		titleFreqSorted = limitToMostFrequent(titleFreq, 5);
		textFreqSorted = limitToMostFrequent(textFreq, 5);
		codeFreqSorted = limitToMostFrequent(codeFreq, 5);
		BufferedWriter writer = new BufferedWriter(new FileWriter("numeric_data.csv"));
		in = new FileReader("merged_data.csv");
		records = CSVFormat.EXCEL.withFirstRecordAsHeader().parse(in);
		writeNumericFeatures(headers, writer, null);
		writeNumericFeatures(tagFreqSorted, writer, null, "tag_", "Tags", stopwords, stemmer);
		writeNumericFeatures(titleFreqSorted, writer, null, "title_", "Title", stopwords, stemmer);
		writeNumericFeatures(textFreqSorted, writer, null, "text_", "ContentText", stopwords, stemmer);
		writeNumericFeatures(codeFreqSorted, writer, null, "code_", "ContentCode", stopwords, stemmer);
		writer.write("\n");
		for (CSVRecord record : records) {
			writeNumericFeatures(headers, writer, record);
			writeNumericFeatures(tagFreqSorted, writer, record, null, "Tags", stopwords, stemmer);
			writeNumericFeatures(titleFreqSorted, writer, record, null, "Title", stopwords, stemmer);
			writeNumericFeatures(textFreqSorted, writer, record, null, "ContentText", stopwords, stemmer);
			writeNumericFeatures(codeFreqSorted, writer, record, null, "ContentCode", stopwords, stemmer);
			writer.write("\n");
		}
		writer.flush();
		writer.close();
	}

	private static Map<String, Integer> limitToMostFrequent(Map<String, Integer> freqSorted, int minimumFreq) {
		Map<String, Integer> mostFreqSorted = new HashMap<String, Integer>();
		for (String token : freqSorted.keySet()) {
			if (freqSorted.get(token) >= minimumFreq) {
				mostFreqSorted.put(token, freqSorted.get(token));
			}
		}
		return mostFreqSorted;
	}

	private static Set<String> loadStopwords(String fileName) throws IOException {
		Set<String> stopwords = new HashSet<String>();
		BufferedReader in = new BufferedReader(new FileReader(fileName));
		String line;
		while ((line = in.readLine()) != null) {
			stopwords.add(line.trim());
		}
		in.close();
		return stopwords;
	}

	private static void writeNumericFeatures(Map<String, Integer> freqSorted, BufferedWriter writer, CSVRecord record,
			String prefix, String columnName, Set<String> stopwords, PorterStemmer stemmer) throws IOException {
		if (record == null) {
			for (String token : freqSorted.keySet()) {
				writer.write(prefix + token + ",");
			}
		} else {
			List<String> tokens = getTokens(record, columnName, stopwords, stemmer);
			for (String token : freqSorted.keySet()) {
				writer.write(countOccurrences(tokens, token) + ",");
			}
		}
	}

	private static List<String> getTokens(CSVRecord record, String columnName, Set<String> stopwords, PorterStemmer stemmer) {
		String string = record.get(columnName);
		if (columnName.equals("Tags")) {
			string = string.replaceAll("><", " ").replaceAll("<", "").replaceAll(">", "");
		}
		string = string.replaceAll("\n", "");
		string = string.replaceAll("[^A-Za-z0-9 ]", "");
		String[] tokens = string.split(" ");
		List<String> tokensList = new ArrayList<String>();
		for (String token : tokens) {
			String trimmedToken = token.toLowerCase().trim();
			if (stopwords.contains(trimmedToken)) {
				continue;
			}
			trimmedToken = stemmer.stem(trimmedToken);
			if (stopwords.contains(trimmedToken)) {
				continue;
			}
			if (trimmedToken.length() > 0) {
				tokensList.add(trimmedToken);
			}
		}
		return tokensList;
	}

	private static void writeNumericFeatures(List<String> headers, BufferedWriter writer, CSVRecord record)
			throws IOException {
		for (String header : headers) {
			if (header.equals("ContentText") || header.equals("ContentCode") || header.equals("Tags")
					|| header.equals("Title") || header.equals("MatchedDeveloperResources")
					|| header.equals("AttractedDeveloperResources") || header.equals("LinkPosition")) {
				continue;
			}
			if (record == null) {
				writer.write(header + ",");
			} else {
				writer.write(record.get(header) + ",");
			}
		}
	}

	private static void updateFrequencies(Map<String, Integer> freq, CSVRecord record, String columnName,
			Set<String> stopwords, PorterStemmer stemmer) {
		List<String> tokens = getTokens(record, columnName, stopwords, stemmer);
		for (String token : tokens) {
			if (!freq.containsKey(token)) {
				freq.put(token, 0);
			}
			freq.put(token, freq.get(token) + 1);
		}
	}

	private static int countOccurrences(List<String> tokens, String item) {
		int counter = 0;
		for (String arrayItem : tokens) {
			if (arrayItem.equals(item)) {
				counter++;
			}
		}
		return counter;
	}

	public static <K, V extends Comparable<? super V>> Map<K, V> sortByValue(Map<K, V> map) {
		List<Entry<K, V>> list = new ArrayList<>(map.entrySet());
		list.sort(Entry.comparingByValue((v1, v2) -> v2.compareTo(v1)));
		Map<K, V> result = new LinkedHashMap<>();
		for (Entry<K, V> entry : list) {
			result.put(entry.getKey(), entry.getValue());
		}
		return result;
	}
}
